import logging
import traceback
from collections.abc import Callable
from functools import wraps
from typing import Any, TypeVar

import dema
import ipyvuetify as v

T = TypeVar("T")


def log_error(func: Callable) -> Callable:
    @wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        try:
            return func(*args, **kwargs)
        except Exception as e:
            func_name = getattr(func, "__name__", "func")
            dema.logger.warning(
                f"Impossible to execute {func_name}, fails with error : {e}"
            )
            for t in traceback.extract_tb(e.__traceback__):
                dema.logger.debug(f"Error in {t.name} at line {t.lineno}: \n\t{t.line}")

    return wrapper


def decorate_all_methods_with_log_error(cls: type) -> None:
    for name, method in cls.__dict__.items():
        if callable(method):
            setattr(cls, name, log_error(method))
        elif type(method) == property:
            decorated_property = property(
                fget=log_error(method.fget) if method.fget is not None else None,
                fset=log_error(method.fset) if method.fset is not None else None,
                fdel=log_error(method.fdel) if method.fdel is not None else None,
            )
            setattr(cls, name, decorated_property)


def log_error_class(original_class: type[T]) -> type[T]:
    # This function is used to decorate each method of a class with log_error
    orig_init = original_class.__init__
    # Make copy of original __init__, so we can call it without recursion

    def __init__(self: type, *args, **kws) -> None:
        decorate_all_methods_with_log_error(original_class)
        log_error(orig_init)(self, *args, **kws)  # Call the original __init__

    original_class.__init__ = __init__  # type: ignore # Set the class' __init__ to the new one

    return original_class


class Logger:
    """
    Each subclass of Logger will automatically have its methods decorated for log_error.

    Without this helper class we would need to decorate each classes with log_error_class.
    Instead we only need to inherite from Logger somehow

    """

    def __new__(cls, *args, **kwargs):
        decorate_all_methods_with_log_error(cls)
        return super().__new__(cls)


class OutputWidgetHandler(logging.Handler):
    """Custom logging handler sending logs to an output widget."""

    def __init__(self, logger_badge: v.Badge, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger_badge = logger_badge
        self.text_area = v.Textarea(
            v_model=None,
            rows=15,
            clearable=True,
            readonly=True,
            solo=True,
            hide_details=True,
        )

    def emit(self, record: logging.LogRecord) -> None:
        """Overload of logging.Handler method."""
        formatted_record = self.format(record) + "\n"
        if self.text_area.v_model is None:
            self.text_area.v_model = ""
        self.text_area.v_model = formatted_record + self.text_area.v_model

        self.logger_badge.v_model = True

        # formatted_record)



from pathlib import Path

import ipyvuetify as v
from dema.front.app import App
from dema.front.waiter import Waiter
from IPython.display import HTML, display

display(HTML(f"<style>{(Path(__file__).parent / 'style.css').read_text()}</style>"))

v.theme.themes.light.primary = "#00205b"
v.theme.themes.dark.primary = "#383838"

# instantiate the app
waiter = Waiter()

__all__ = ["App"]



from __future__ import annotations

import polars as pl
from polars import col as c

from dema import DataEngine

def get_concept_columns_repr(
    engine: DataEngine, 
    concept: str
) -> dict[str, pl.LazyFrame]:
    columns_repr = {}
    for column, fk_concept in engine.get_concept_desc(concept).select('column', 'fk_concept').iter_rows():
        if fk_concept:
            fk_concept_desc = engine.get_concept_desc(fk_concept) 
            fk_column = fk_concept_desc.filter(c.type == 'primary')['column'].item()

            cols_to_display = fk_concept_desc.filter(c.type.is_not_null())['column'].to_list()
            prefix = column[: len(column) - len(fk_column)]
            columns_repr[column] = (
                engine.read_concept(fk_concept)
                .select(
                    fk_column, pl.concat_str(cols_to_display, separator=" - ").suffix("__repr")
                )
                .select(pl.all().name.prefix(prefix))
            )
    return columns_repr



import dema.front


class Waiter:
    def __enter__(self):
        dema.front.app.linear_progess.show()
        dema.front.app.logger_bottom_sheet.v_model = True
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):  # noqa: ANN001
        dema.front.app.linear_progess.hide()
        if exception_type is None:
            dema.front.app.logger_bottom_sheet.v_model = False
            dema.front.app.app_bar.logger_badge.v_model = False



# this files might need to be moved to scema SCenario Engine Manager

from typing import Any
from dema.engine import DataEngine
from dema.front.manager.concept import EditingConceptManager
from ipyvuetable.widgets import FileInput
import polars as pl
import ipywidgets as widgets
import ipyvuetify as v


class SupersetManager(EditingConceptManager):
    # list of columns that don't trigger a recompute of the children
    ignore_changes_columns: list[str] = ["id", "name"]

    # list of concepts attached to this superset
    # they we will displayed below the superset table
    children_concepts: list[str] = []

    def __init__(self, engine: DataEngine, concept: str, **kwargs: Any):
        super().__init__(engine, concept=concept + "_superset", **kwargs)
        self.children_content = v.Flex()

        self.ui = v.Flex(children = [self.ui, self.children_content])

    def _on_save_dialog(self, *args):
        super()._on_save_dialog(*args)

        # first get the id to update based on diffenrence between new_items and previous_items
        id_to_updates = []
        for item in self.new_items:
            id_ = item[self.item_key]
            previous_item = self.previous_items.get(id_, {})
            for column in item:
                if column not in self.ignore_changes_columns:
                    if item[column] != previous_item.get(column):  # type: ignore
                        id_to_updates.append(id_)
                        continue

        # update all ids
        if id_to_updates:
            df_dict = self._load_dataframes_from_file_inputs()

            # let the opportunity to a children classe to overwrite
            df_dict_preprocessed = self.preprocess_inputs(**df_dict)

            for concept_id in id_to_updates:
                for concept, df in df_dict_preprocessed.items():
                    self.engine.to_concept(df, concept, concept_id)

    def preprocess_inputs(self, **df_dict: pl.DataFrame) -> dict[str, pl.DataFrame]:
        raise NotImplementedError(
            "preprocess_inputs shoud be defined by children class"
        )

    def _load_dataframes_from_file_inputs(self) -> dict[str, pl.DataFrame]:
        return {
            column: pl.concat(widget.load_dataframes())
            for column, widget in self.dialog_widgets.items()
            if isinstance(widget, FileInput)
        }

    def on_nb_selected(self, change: dict[str, Any]):
        super().on_nb_selected()
        if len(self.selected_keys) == 1:
            tab_list = [v.Tab(children=[c]) for c in self.children_concepts]

            content_list = [
                v.TabItem(
                    children=[
                        EditingConceptManager(
                            self.engine,
                            concept=tab.children[0],
                            concept_id=self.selected_keys[0],
                        ).ui
                    ]
                )
                for tab in tab_list
            ]

            self.children_content.children = [
                v.Tabs(v_model=0, children=tab_list + content_list, color="primary")
            ]
        else:
            self.children_content.children = []


    def _on_click_delete_btn(self, *args):
        ids_to_delete = self.selected_keys
        # we delete it parent first
        super()._on_click_delete_btn()

        print(ids_to_delete)
        self.engine.delete_concept(
            concept=self.children_concepts,
            concept_ids=ids_to_delete

        )


from functools import cached_property
import os
from pathlib import Path

import polars as pl
from polars import col as c
import dema
from dema.back.exception import NotInConceptDesc

import dema.back.utils_io as utils_io
from dema.back import utils_sql
from dema.database import LogicalDataHash


class DataEngine:
    def __init__(
        self,
        name: str,
        env: str = "dev",
        data_root_path: Path | None = None,
        concepts_desc_path: Path | None = None,
        front_structure_path: Path | None = None,
        dev_mode=True,
    ) -> None:
        self.name = name
        self.env = env
        self.dev_mode = dev_mode

        self.data_root_path = (
            (
                Path(os.getenv("DEMA_DATA_ROOT_PATH", "~/.dema"))
                if data_root_path is None
                else data_root_path
            )
            .expanduser()
            .absolute()
        )

        self.physical_data_path = self.data_root_path / self.name / "hashes"
        self.logical_data_path = self.data_root_path / self.name / self.env / "data"
        self.db_path = self.data_root_path / self.name / self.env / "database.db"

        utils_io.make_dirs(self.physical_data_path, self.logical_data_path)

        self.front_structure_path = front_structure_path
        self.concepts_desc_path = concepts_desc_path or self.logical_data_path / "concepts_desc.csv"

        # self.concepts_desc = utils_io.read_descriptors(self.concepts_desc_path)
        self.sql_engine = utils_sql.get_sql_engine(self.db_path)

        self.to_concept(self.concepts_desc, 'concepts_desc')

    @cached_property
    def engine_concepts_desc(self) -> pl.DataFrame:
        return pl.read_csv(
            f"{dema.__engine_root_path__}/concepts_desc.csv",
            schema={
                "concept": pl.String,
                "column": pl.String,
                "data_type": pl.String,
                "type": pl.String,
                "null_forbidden": pl.Boolean,
                "fk_concept": pl.String,
            },
        )
    
    @property
    def concepts_desc(self) -> pl.DataFrame:
        if not hasattr(self, '_concepts_desc'):

            if self.concepts_desc_path.exists():
                self._concepts_desc = pl.read_csv(
                    self.concepts_desc_path,
                    schema=self.engine_concepts_desc.schema,
                )
            else: 
                self._concepts_desc = pl.DataFrame(schema=self.engine_concepts_desc.schema)
                
        return self._concepts_desc
    
    @concepts_desc.setter
    def concepts_desc(self, df: pl.DataFrame):
        self._concepts_desc = df

        dema.logger.debug(f'writing {self.concepts_desc_path}')
        df.write_csv(self.concepts_desc_path)


    def read_concept(
        self,
        concept: str,
        concept_id: int | None = None,
        version: int | None = 0,
    ) -> pl.LazyFrame:
        """Return a LazyFrame of the concept."""
        if version == 0:  # then we can use symlink
            path = utils_io.get_logical_path(self.logical_data_path, concept, concept_id)
            paths = [path] if path.exists() else []
        else:
            paths = [
                self.physical_data_path / f"{h}.parquet"
                for h in self.query_logical_data_hash(
                    columns=["hash"],
                    version=version,
                    concept=concept,
                    concept_id=concept_id,
                )
                .to_series()
                .to_list()
            ]
        match len(paths):
            case 0:
                concept_desc = self.get_concept_desc(concept)
                df = pl.LazyFrame(schema=utils_io.get_polars_schema(concept_desc))
            case 1:
                df = pl.scan_parquet(paths[0])
            case _:
                df = pl.scan_parquet(paths)

        return df


    def query_logical_data_hash(
        self,
        concept: list[str] | str | None = None,
        concept_id: list[int] | int | None = None,
        columns: list[str] | None = None,
        version: int | None = 0,
    ) -> pl.DataFrame:
        """Return the rows of LogicalDataHash for this concept, concept_id, version"""
        where = utils_io.dict_to_sql_where_statement(
            {
                "concept": concept,
                "concept_id": concept_id,
            }
        )

        query = "\n".join(
            [
                f"select {', '.join(columns) if columns else '*'} from (",
                "    select *,",
                "    -RANK() OVER ("
                "        PARTITION BY concept, concept_id ORDER BY timestamp_ns desc"
                "    ) + 1 AS version",
                "    FROM logical_data_hash",
                f"    {where}",
                ")",
                utils_io.dict_to_sql_where_statement({"version": version}),
            ]
        )
        model_json_schema = LogicalDataHash.model_json_schema()
        df = utils_io.read_from_sqlite(
            query,
            self.sql_engine,
            schema_overrides=utils_io.get_db_table_schema(model_json_schema),
        ).filter(c.hash.is_not_null())

        return df

    def to_concept(
        self,
        df: pl.LazyFrame | pl.DataFrame,
        concept: str,
        concept_id: int | None = None,
    ) -> None:
        """Save a concept in parquet if it complies with descriptor rules."""

        dema.logger.debug(f"to_concept: {concept}, {concept_id}")
        concept_desc = self.get_concept_desc(concept)

        # make sure schema is correct
        schema = utils_io.get_polars_schema(concept_desc)
        df = df.lazy().select(schema.keys()).cast(schema).collect()  # type: ignore

        # check pk validity
        if pk := utils_io.get_pks(concept_desc):
            assert_msg = f"Primary keys ({pk}) are not unique for {concept}"
            assert df.select(pk).is_unique().all(), assert_msg

        hash_ = utils_io.hash_dataframe(df)

        # save to parquet
        utils_io.write_hash(self.physical_data_path, df, hash_)

        # add line to logical_data_hash
        self.append_to_logical_data_hash(
            pl.DataFrame(
                {
                    "concept": concept,
                    "concept_id": concept_id,
                    "hash": hash_,
                }
            )
        )
        if concept == 'concepts_desc':
            self.concepts_desc = df
            
    def delete_concept(
        self,
        concept: list[str] | str | None = None,
        concept_ids: list[int] | None = None,
    ) -> None:
        """Logical deletion of this concept."""
        data_hash = self.query_logical_data_hash(concept, concept_ids)

        self.append_to_logical_data_hash(
            data_hash.with_columns(pl.lit(None, dtype=pl.Utf8).alias("hash")),
        )

    def get_concept_desc(self, concept: str) -> pl.DataFrame:
        concept_desc = self.concepts_desc.filter(c.concept == concept)
        if concept_desc.is_empty():
            concept_desc = self.engine_concepts_desc.filter(c.concept == concept)
            if concept_desc.is_empty():
                raise NotInConceptDesc(f"{concept} is not a concept.")
        return concept_desc
    
    def append_to_logical_data_hash(self, data_hash: pl.DataFrame) -> None:
        # remove computed by sqlmodel
        data_hash = data_hash.drop("id", "timestamp_ns")
        rows = utils_sql.add_rows(
            self.sql_engine,
            LogicalDataHash,
            data_hash.rows(named=True),
        )
        utils_io.update_logical_paths(
            self.logical_data_path, self.physical_data_path, rows
        )



import logging as _logging
from pathlib import Path
from dema.engine import DataEngine

__engine_root_path__ = Path(__file__).parent
__version__ = "0.2.5"

# set the logger
logger = _logging.getLogger("dema")
logger.setLevel("DEBUG")
file_handler = _logging.FileHandler(filename=__engine_root_path__ / 'logs')
file_handler.setFormatter(_logging.Formatter('%(asctime)s - [%(levelname)s] %(message)s', "%Y-%m-%d %H:%M:%S"))
logger.addHandler(file_handler)

__all__ = ["DataEngine"]



from pathlib import Path
from typing import Any, Sequence, TypeVar

from sqlalchemy.future import Engine
from sqlmodel import Session, SQLModel, and_, create_engine, select
from sqlmodel.sql.expression import SelectOfScalar

T = TypeVar("T", bound=SQLModel)


def get_sql_engine(path: Path) -> Engine:
    """return an SQLite engine and create the database if it doesn't exist"""
    sql_engine = create_engine(f"sqlite:///{path}")
    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
        SQLModel.metadata.create_all(sql_engine)
    return sql_engine


def select_rows(engine: Engine, table: type[T], **filters: Sequence) -> list[T]:
    """Filter `table`"""
    with Session(engine) as session:
        filters_statement = [getattr(table, column).in_(values) for column, values in filters.items()]
        statement = select(table).where(and_(*filters_statement))
        rows = list(session.exec(statement).all())
    return rows


def add_rows(
    engine: Engine,
    table: type[T],
    data: Sequence[dict[str, Any]],
) -> list[T]:
    """Add rows to `table`"""
    with Session(engine) as session:
        rows = [table.model_validate(d) for d in data]
        session.add_all(rows)

        # commit session and refresh
        session.commit()
        for row in rows:
            session.refresh(row)
    return rows


def _filter_on_primary_keys(
    table: type[T],
    record: dict[str, Any],
    on: list[str],
) -> SelectOfScalar[T]:
    """Return a sqlmodel where statement with rows that match `records`"""
    match_on_pks = [getattr(table, column) == record[column] for column in on]
    statement = select(table).where(and_(*match_on_pks))
    return statement


def upsert_rows(
    engine: Engine, table: type[T], data: list[dict[str, Any]], on: list[str]
) -> list[T]:
    """Upsert `table` with records in `data` base on primary keys `on`."""

    with Session(engine) as session:
        rows = []
        for record in data:
            statement = _filter_on_primary_keys(table, record, on)

            # update row or create new
            if row := session.exec(statement).first():
                for col, value in record.items():
                    setattr(row, col, value)
            else:
                row = table.model_validate(record)

            # add to the session
            session.add(row)
            rows.append(row)

        # commit session and refresh
        session.commit()
        for row in rows:
            session.refresh(row)

        return rows


def delete_rows(
    engine: Engine,
    table: type[SQLModel],
    data: Sequence[dict[str, Any]],
    on: list[str],
) -> None:
    """Delete rows in `table` based"""
    with Session(engine) as session:
        for record in data:
            statement = _filter_on_primary_keys(table, record, on)
            session.delete(statement)
        session.commit()




import polars as pl

from dema.engine import DataEngine

def infer_primary_keys(df: pl.DataFrame) -> list[str]:
    columns = df.columns
    for i in range(len(columns)):
        if df.select(columns[:i + 1]).is_unique().all():
            return columns[:i + 1]
    raise Exception('No primary keys detected')
    
def update_concepts_desc(
    engine: DataEngine,
    df: pl.DataFrame, 
    concept: str
) -> None:
    pks = infer_primary_keys(df)
    new_concepts_desc= pl.DataFrame([
        dict(
            concept = concept,
            column = column,
            data_type = data_type,
            type = 'primary' if column in pks else None,
            null_forbidden = df[column].is_not_null().all(),
            fk_concept = None,
        ) for column, data_type in zip(df.columns, df._df.dtype_strings())
    ], schema=engine.engine_concepts_desc.schema)

    engine.concepts_desc = pl.concat([
        engine.concepts_desc.update(new_concepts_desc, on = ['concept', 'column'], how = 'left', include_nulls=True),
        new_concepts_desc.join(engine.concepts_desc, on = ['concept', 'column'], how = 'anti')
    ])


from __future__ import annotations
import collections.abc

import functools
import hashlib
from pathlib import Path
from typing import TYPE_CHECKING, Any

import polars as pl
from polars import col as c
from polars.datatypes.convert import dtype_short_repr_to_dtype
from sqlalchemy import Engine

import dema


if TYPE_CHECKING:
    from polars.type_aliases import SchemaDict

    from dema.database import LogicalDataHash

PYDANTIC_TO_POLARS_TYPE = {
    "integer": pl.Int32,
    "number": pl.Int64,
    "string": pl.String,
    "date-time": pl.Datetime,
    "float": pl.Float64,
    "array[string]": pl.List(pl.String),
    "array[integer]": pl.List(pl.Int32),
    None: pl.Null,
}

def make_dirs(physical_data_path: Path, logical_data_path: Path):
    """
    the file structure is the following:
    ~/.dema                                 <--- root_data_path
    └── name
        ├── env
        │   ├── data                        <--- logical_data_path
        │   │   ├── table1
        │   │   │   └── 1.parquet
        │   │   │   └── 2.parquet        
        │   │   ├── table2
        │   │   │   └── 1.parquet
        │   └── database.sql
        ├── hashes                           <--- physical_data_path
        │   ├── alknfskfmanfzafralh.parquet
        │   └── dlkne6dnl91n3bqsjfg.parquet
        │   └── fskfmanfzazn3bqfska.parquet
    """

    logical_data_path.mkdir(exist_ok=True, parents=True)
    physical_data_path.mkdir(exist_ok=True, parents=True)


def get_logical_path(
    logical_data_path: Path,
    concept: str,
    concept_id: int | None = None,
) -> Path:
    """
    Return list a path for the given concept
    """
    return logical_data_path / concept / f"{concept_id or '*'}.parquet"


def update_logical_paths(
    logical_data_path: Path,
    physical_data_path: Path,
    logical_data_hashes: list[LogicalDataHash],
) -> None:

    for logical_data_hash in logical_data_hashes:
        logical_path = get_logical_path(
            logical_data_path, logical_data_hash.concept, logical_data_hash.concept_id
        )
        # new is inexistent => remove
        if logical_data_hash.hash is None:
            if logical_path.exists():
                dema.logger.debug(f"unlink {logical_path}")
                logical_path.unlink()
            continue

        # new is the same => do nothing
        if logical_path == logical_path.resolve().stem:
            continue
    
        # else update
        if logical_path.exists():
            logical_path.unlink()
        dema.logger.debug(f"update symlinks {logical_path}")
        logical_path.parent.mkdir(parents=True, exist_ok=True)
        logical_path.symlink_to(
            physical_data_path.absolute() / f"{logical_data_hash.hash}.parquet"
        )


def write_hash(path: Path, df: pl.DataFrame, hash_: str) -> None:
    path = path / f"{hash_}.parquet"
    if not path.exists():
        df.write_parquet(path)


def hash_dataframe(df: pl.DataFrame) -> str:
    values_hash = str(df.hash_rows().sort().implode().hash().item())
    schema_hash = str(df.schema)

    return hashlib.md5((values_hash + schema_hash).encode()).hexdigest()


@functools.cache
def str_to_polars_dtype(dtype_str: str) -> pl.PolarsDataType:
    dtype = dtype_short_repr_to_dtype(dtype_str)
    assert dtype is not None, f"impossible to parse {dtype_str}"
    return dtype


def get_polars_schema(concept_desc: pl.DataFrame) -> SchemaDict:
    schema = {
        col: str_to_polars_dtype(dtype)
        for col, dtype in concept_desc.select('column', 'data_type').iter_rows()
    }
    return schema


def get_pks(concept_desc: pl.DataFrame) -> list[str]:
    """Return primary keys of a concept."""
    return concept_desc.filter(c.type == 'primary')['column'].to_list()


def read_from_sqlite(
    query: str, engine: Engine, schema_overrides: SchemaDict
) -> pl.DataFrame:
    """
    sqlite stores list and datetime as strings
    so we need to manually cast the type
    """
    list_columns = [k for k, v in schema_overrides.items() if v == pl.List]
    datetime_columns = [k for k, v in schema_overrides.items() if v == pl.Datetime]

    df = (
        pl.read_database(
            query,
            engine,
            schema_overrides={
                k: v if k not in (*list_columns, *datetime_columns) else pl.String
                for k, v in schema_overrides.items()
            },
        )
        .with_columns(
            pl.col(list_columns).str.json_decode(),
            pl.col(datetime_columns).str.to_datetime(),
        )
        # https://github.com/pola-rs/polars/issues/14468
        .cast(schema_overrides)  # type: ignore
    )
    return df


def get_db_table_schema(
    model_json_schema: dict[str, Any],
    return_string: bool = False
) -> dict[str, Any]:
    """Return a polars schema from a SQLModel table"""

    def _parse_type(prop: dict[str, Any]) -> str:
        # Literal case
        if "$ref" in prop:
            ref = prop["$ref"].split("/")[-1]
            type_ = model_json_schema["$defs"][ref]["type"]
        elif "type" in prop:
            if prop["type"] == "array":
                type_ = f"{prop['type']}[{prop['items']['type']}]"
            elif prop.get("format") == "date-time":
                type_ = "date-time"
            else:
                type_ = prop["type"]
        else:
            raise Exception(f"Type {prop} is not supported")
        return type_

    polars_schema = {}
    for field, prop in model_json_schema["properties"].items():
        type_: str | None
            
        if "anyOf" in prop:
            type_ = next(
                iter(
                    _parse_type(inner_prop)
                    for inner_prop in prop["anyOf"]
                    if inner_prop.get("type") != "null"
                ),
                None,
            )
        else:
            type_ = _parse_type(prop)

        if not return_string:
            polars_schema[field] = PYDANTIC_TO_POLARS_TYPE[type_]
    return polars_schema  # type: ignore[return-value]


def dict_to_sql_where_statement(filters: dict[str, Any]) -> str:
    """Parse a dictionnary into a sql where statement.
    Examples
    --------
    >>> print(utils_io.dict_to_where_closure({'type': 'SRC', 'concept': ['AC_MODEL', 'AC_MANUF']}))
    where
    type = 'SRC' and concept in ('AC_MODEL', 'AC_MANUF')
    """
    constraints: list[str] = []
    for column, value in filters.items():
        if value is not None:
            if isinstance(value, collections.abc.Sequence) and not isinstance(
                value, str
            ):
                if len(value) == 1:
                    constraints.append(f"{column} = {next(iter(value))!r}")
                elif len(value) > 1:
                    constraints.append(f"{column} in {tuple(value)!s}")
            else:
                constraints.append(f"{column} = {value!r}")

    where = "where\n" + " and ".join(constraints) if constraints else ""
    return where



import json
import logging
from pathlib import Path
from typing import Any, Self
from ipywidgets import jsdlink

import dema
import ipyvuetify as v
import traitlets as t
from dema.engine import DataEngine
from dema.front.logger import Logger, OutputWidgetHandler
from dema.utils.utils_misc import import_class


class App(v.App, Logger):
    def __init__(self, *, engine: DataEngine, app: bool | None = True):
        self.app = app
        self.engine = engine
        self.app_bar = AppBar(engine=engine, app=app)
        self.tree_view = TreeView(engine.front_structure_path)
        self.navigation_drawer = v.NavigationDrawer(
            app=app, v_model=True, clipped=True, children=[self.tree_view]
        )
        self.content = v.Content()  # class_="px-1 py-0"

        handler = OutputWidgetHandler(self.app_bar.logger_badge)
        handler.setFormatter(
            logging.Formatter("%(asctime)s - [%(levelname)s] %(message)s", "%H:%M:%S")
        )
        dema.logger.addHandler(handler)

        self.logger_w = handler.text_area
        self.linear_progess = v.ProgressLinear(
            color="primary", indeterminate=True, class_="d-none"
        )
        self.logger_bottom_sheet = v.BottomSheet(
            v_model=False, children=[self.linear_progess, self.logger_w]
        )

        super().__init__(
            children=[
                self.app_bar,
                self.navigation_drawer,
                self.content,
                self.logger_bottom_sheet,
            ]
        )

        self.logger_w.observe(self.on_change_text_area, names="v_model")
        self.app_bar.logger_icon.on_event("click", self.toggle_logger)
        self.app_bar.nav_icon.on_event("click", self.toggle_navigation_drawer)
        self.tree_view.observe(self.on_click_treeview, names="activated")
        self.logger_w.on_event("click:clear", self._on_click_clear_text_area)

    def _on_click_clear_text_area(self, *args) -> None:
        self.logger_bottom_sheet.v_model = False

    def toggle_logger(self, *args) -> None:
        # self.logger_bottom_sheet.class_list.toggle('d-none')
        self.logger_bottom_sheet.v_model = not self.logger_bottom_sheet.v_model

        # reset log counter
        self.app_bar.logger_badge.v_model = False

    def toggle_navigation_drawer(self, *args) -> None:
        self.navigation_drawer.v_model = not self.navigation_drawer.v_model

        # usefull only in app=False aka debug
        if not self.app:
            if self.navigation_drawer.v_model:
                self.navigation_drawer.show()
            else:
                self.navigation_drawer.hide()

    def on_click_treeview(self, change: dict) -> None:
        new = change["new"]
        if new:
            if new[0] is not None:
                class_path = new[0]["class"]
                args = new[0].get("args", [])
                kwargs = new[0].get("kwargs", {})
                class_instance = import_class(class_path)(self.engine, *args, **kwargs)
                self.content.children = [
                    class_instance.ui if hasattr(class_instance, "ui") else class_instance
                ]
        else:
            self.content.children = []

    def on_change_text_area(self, change: dict) -> None:
        self.app_bar.logger_icon.disabled = not bool(change.get("new"))

    def display_in_dialog(self, *, v_model: bool = True):
        # wrap the App instance inside a Dialog in fullscreen to be used inside jupyter
        dialog = v.Dialog(
            v_model=v_model,
            fullscreen=True,
            persistent=True,
            no_click_animation=True,
            retain_focus=False,
            v_slots=[
                {
                    "name": "activator",
                    "variable": "x",
                    "children": v.Btn(v_on="x.on", children=["Open App"]),
                }
            ],
            children=[v.Card(children=[self])],
        )

        def toggleLoading(*args) -> None:
            dialog.v_model = not dialog.v_model

        dialog.on_event("keydown.esc", toggleLoading)

        return dialog


class AppBar(v.AppBar, Logger):
    def __init__(self, *, engine: DataEngine, app: bool | None = False):
        self.engine = engine
        self.nav_icon = v.AppBarNavIcon()
        self.logger_icon = v.Icon(children=["mdi-book-open-outline"])
        self.logger_badge = v.Badge(
            bottom=True,
            v_model=False,
            color="red",
            v_slots=[
                {
                    "name": "badge",
                    "children": ["!"],
                }
            ],
            children=[self.logger_icon],
            class_="mx-5",
        )
        self.search = v.Combobox(
            v_model=None,
            item_value="class",
            label=f"Explore {engine.name} ...",
            rounded=True,
            clearable=True,
            single_line=True,
            light=True,
            background_color="white",
            class_="mt-5",
        )

        self.left_btn = v.Icon(
            children=["mdi-arrow-left"], icon=True, disabled=True, class_="pl-2"
        )
        self.right_btn = v.Icon(
            children=["mdi-arrow-right"], icon=True, disabled=True, class_="pl-2"
        )
        self.reload_btn = v.Icon(
            children=["mdi-reload"], icon=True, disabled=True, class_="pl-2"
        )

        version_chip = v.Chip(
            children=[f"v.{dema.__version__}"], outlined=True, class_="ml-2"
        )
        env_chip = v.Chip(children=[self.engine.env], outlined=True, class_="ml-2")

        super().__init__(
            app=app,
            dark=True,
            clipped_left=True,
            clipped_right=True,
            color="primary",
            children=[
                self.nav_icon,
                v.ToolbarTitle(children=[engine.name]),
                v.Spacer(),
                self.search,
                v.Col(
                    children=[
                        v.Row(
                            style_="flex-wrap:nowrap",
                            children=[self.left_btn, self.right_btn, self.reload_btn],
                        )
                    ]
                ),
                self.logger_badge,
                version_chip,
                env_chip,
            ],
        )


class TreeView(v.Treeview, Logger):
    # active is a trait from v.Treeview, event from the front are handled with `update:active`
    # but self.active is not sync on the back (BUG ??)
    # The workaound is to create another trait: `activated` that is jsdlink with `active`
    activated = t.List(t.Any(), default_value=None, allow_none=True).tag(sync=True)
    
    def __init__(self, treeview_path: Path | None):
        if treeview_path and treeview_path.exists():
            items = json.loads(treeview_path.read_text())
        else:
            items = []

        # this is used to need which node get opened
        self.previous_open_ids: list[str] = []
        super().__init__(
            items=items,
            dense=True,
            open_on_click=True,
            activatable=True,
            transition=True,
            return_object=True,
        )

        def _on_update_active(
            widget: Self, event: str, data: list[Any | None]
        ) -> None:
            # there is a bug `active is not properly sync`
            widget.__dict__["_trait_values"]["active"] = data
            widget.activated = data if data else []

        jsdlink((self, "active"), (self, "activated"))
        self.on_event("update:active", _on_update_active)



import time
from datetime import datetime

from sqlalchemy.orm import declarative_base
from sqlmodel import Field, SQLModel

Base = declarative_base()


class LogicalDataHash(SQLModel, table=True):
    __tablename__: str = "logical_data_hash"

    id: int | None = Field(primary_key=True, default=None)
    concept: str = Field(index=True)
    concept_id: int | None = Field(ge=1, index=True, default=None, nullable=True)
    timestamp_ns: datetime = Field(default_factory=datetime.utcnow)
    hash: str | None


class BricksComputation(SQLModel, table=True):
    __tablename__: str = "bricks_computation"

    inputs_data_hash_list: str = Field(primary_key=True)
    brick_hash: str = Field(primary_key=True)
    outputs_data_hash_list: str
    computation_time_s: float = Field(gt=0)
    timestamp_ns: datetime = Field(default_factory=time.time_ns)


from dema.front.logger import Logger, log_error_class # type: ignore
from ipyvuetable import EditingTable, Table

# def log_error_class(d):  # noqa: F811
#     return d  # type: ignore

class TableManager(log_error_class(Table), Logger):
    ...


class EditingTableManager(log_error_class(EditingTable), Logger):
    ...



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dema",
    version="0.2.5",
    author="Gabriel Robin",
    description="Data Engine MAnagment tool for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gab23r/dema",
    packages=setuptools.find_packages(),
    install_requires=["sqlmodel", "polars"],
    extras_require={
        "front": ["ipyvuetable", "ipyevents"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={"": ["front/style.css", "concepts_desc.csv"]},
    python_requires=">=3.11",
)



from typing import Any

import dema.back.utils_io as utils_io
import polars as pl
from dema.back.utils_view import get_concept_columns_repr
from dema.engine import DataEngine
from dema.front.manager.base import EditingTableManager, TableManager
from polars.type_aliases import IntoExpr


class EditingConceptManager(EditingTableManager):
    def __init__(self, engine: DataEngine, concept: str, concept_id: int | None = None, **kwargs: Any):
        self.engine = engine
        self.concept = concept
        self.concept_id = concept_id
        self.concept_desc = engine.get_concept_desc(concept)
        df=self.get_df()

        # if the primary key is unique and is interger
        # we auto increment it on new items
        self.auto_increment_key = False
        if len(pks := utils_io.get_pks(self.concept_desc)) == 1:
            kwargs['item_key'] = pks[0]
            self.auto_increment_key = df.schema.get(pks[0]) in pl.INTEGER_DTYPES

        if self.auto_increment_key:
            kwargs['hide_dialog_keys'] = kwargs.get('hide_dialog_keys', [])
            kwargs['hide_dialog_keys'].append(kwargs['item_key'])

        # default kwargs
        # kwargs['class_'] = kwargs.get('class_', 'extra_dense')
        kwargs["show_select"] = kwargs.get("show_select", True)
        kwargs["title"] = kwargs.get("title", concept)

        super().__init__(
            df=df, 
            columns_repr=get_concept_columns_repr(engine, concept), 
            **kwargs
        )


    def get_df(self) -> pl.LazyFrame:
        return self.engine.read_concept(self.concept, self.concept_id)

    def get_default_new_item(self) -> dict[str, IntoExpr]:
        default_new_item = {}

        # if pk is int, we increment it by one-
        if self.auto_increment_key:
            default_new_item[self.item_key] = (
                self.df.select(pl.col(self.item_key).fill_null(pl.col(self.item_key).max().fill_null(0) + 1)).collect().item()
            )

        return default_new_item

    def on_df_change(self) -> None:
        super().on_df_change()
        self.engine.to_concept(self.df, self.concept)


class ConceptManager(TableManager):
    def __init__(self, engine: DataEngine, concept: str, concept_id: int | None, **kwargs: Any):
        self.engine = engine
        self.concept = concept
        df = self.engine.read_concept(concept)
        columns_repr = get_concept_columns_repr(engine, concept)

        # default kwargs
        # kwargs['class_'] = kwargs.get('class_', 'extra_dense')
        kwargs["show_select"] = kwargs.get("show_select", True)

        super().__init__(df=df, title=concept, columns_repr=columns_repr, **kwargs)


    def on_df_change(self) -> None:
        super().on_df_change()
        self.engine.to_concept(self.df, self.concept)




import importlib
import time
from collections.abc import Callable
from functools import wraps
from typing import Any

import dema


def timeit(method: Callable) -> Callable:
    @wraps(method)
    def timed(*args, **kwargs) -> Any:
        t0 = time.perf_counter()
        result = method(*args, **kwargs)
        t1 = time.perf_counter()
        dema.logger.debug(f"{method.__name__} {(t1 - t0) * 1000:2.2f} ms")
        return result

    return timed


def import_class(path: str) -> Any:
    package, module_name, class_name = path.rsplit(".", maxsplit=2)

    # import the module
    module = importlib.import_module(f".{module_name}", package=package)

    return module.__dict__[class_name]
